# PWBus - Message Class
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Sat Nov 16 16:36:36 -03 2019

import json
import traceback

from pwbus.commons.logging import *

# _Message
#
#


class _Message:

    PAYLOAD_TYPES = [
        {"type": "text",    "content-type": "text/plain"},
        {"type": "json",    "content-type": "application/json"},
        {"type": "xml",     "content-type": "application/xml"},
        {"type": "iso8583", "content-type": "text/plain"},
        {"type": "csv",     "content-type": "text/csv"},
    ]

    def __init__(self):
        self.payload_format = "json"
        self.task_id = None
        self.message_id = None
        self.correlation_id = None
        self.task_id = None
        self.headers = dict()
        self.payload = dict()
        self.payloadAsString = None

    # _Message.setPayloadFormat
    #
    def setPayloadFormat(self, payload_format):
        self.payload_format = payload_format

    # _Message.getMessageId
    #
    def getMessageId(self):
        return self.message_id

    # _Message.setMessageId
    #
    def setMessageId(self, message_id):
        self.message_id = message_id

    # _Message.getCorrelationId
    #
    def getCorrelationId(self):
        return self.correlation_id

    # _Message.setCorrelationId
    #
    def setCorrelationId(self, correlation_id):
        self.correlation_id = correlation_id

    # _Message.getTaskId
    #
    def getTaskId(self):
        return self.task_id

    # _Message.setTaskId
    #
    def setTaskId(self, task_id):
        self.task_id = task_id

    # _Message.getPayload
    #
    def getPayload(self):
        return self.payload

    # _Message.getHeaders
    #
    def getHeaders(self):
        return self.headers

    # _Message.setHeaders
    #
    def setHeaders(self, headers):
        self.headers = headers

    # _Message.setHeadersEntry
    #
    def setHeadersEntry(self, name, value):
        self.headers[name] = value

    # _Message.getHeadersEntry
    #
    def getHeadersEntry(self, name):
        return self.headers[name] if name in self.headers else None

    # _Message.setPayload
    #
    def setPayload(self, payload, create_correlation_id=False, keep_correlation_id=True):
        if not isinstance(payload, dict):
            log_debug(
                "_Message.setPayload - payload is invalid, need be 'dict' type")
            return None
        self.payload = payload

    # _Message.setPayloadField
    #
    def setPayloadField(self, name, value):
        self.payload[name] = value

    # _Message.getPayloadStream
    #
    def getPayloadStream(self):
        return self.getPayloadAsString().encode('utf-8')

    # _Message.setPayloadAsString
    #
    def setPayloadAsString(self, string):
        self.payloadAsString = string

    # _Message.getPayloadAsString
    #
    def getPayloadAsString(self):
        if self.payloadAsString:
            return self.payloadAsString

        log_debug("_Message.getPayloadAsString type(payload)",
                  type(self.payload))
        if isinstance(self.payload, dict):
            if 'payload' in self.payload:
                return json.dumps(self.payload['payload'])
            else:
                return json.dumps(self.payload)

        log_debug("_Message.getPayloadAsString type(payload)",
                  type(self.payload))
        return self.payload



# PWBus - Client Class
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Sat Dec 14 19:27:34 -03 2019

# Client
#
#


class Client:

    # Client.clear_header
    #
    def clear_header(self, response):
        for field in list(response):
            if field.lower().startswith('pwbus-'):
                del response[field]
        return response



# PWBus - Flow Class
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Sat Nov 16 16:36:36 -03 2019

import secrets
from pwbus.flows._message import _Message

# _Flow
#
#


class _Flow:

    ALLOWED_HEADERS = [
        "Pwbus-Channel",
        "Pwbus-Message-Id",
        "Pwbus-Correlation-Id",
        "Pwbus-Task-Id"
    ]

    def __init__(self, flow_type, connector):
        self.flow_type = flow_type
        self.connector = connector
        self.message = _Message()
        self.response = None
        self.registry = None
        self.running = True

    # _Flow.getType
    #
    def getType(self):
        return self.flow_type

    # _Flow.getConnector
    #
    def getConnector(self):
        return self.connector

    """ MESSAGE METHODS """

    # _Flow.getMessage
    #
    def getMessage(self):
        return self.message

    # _Flow.setMessage
    #
    def setMessage(self, message):
        self.message = message

    """ PAYLOAD METHODS """

    # _Flow.getPayloadField
    #
    def getPayloadField(self, name):
        return self.message.getPayloadField(name)

    # _Flow.setPayloadField
    #
    def setPayloadField(self, name, value):
        self.message.setPayloadField(name, value)

    """ HEADERS METHODS """

    # _Flow.getHeaders
    #
    def getHeaders(self):
        return self.message.getHeaders()

    # _Flow.setHeaders
    #
    def setHeaders(self, headers):
        self.message.setHeaders(headers)

    # _Flow.getHeadersEntry
    #
    def getHeadersEntry(self, name):
        return self.message.getHeadersEntry(name)

    # _Flow.setHeadersEntry
    #
    def setHeadersEntry(self, name, value):
        self.message.setHeadersEntry(name, value)

    """ RESPONSE METHODS """

    # _Flow.getResponse
    #
    def getResponse(self, field=None):
        if field:
            return self.response[field]
        else:
            return self.response

    # _Flow.setResponse
    #
    def setResponse(self, response):
        self.response = response

    # _Flow.setResponseEntry
    #
    def setResponseEntry(self, key, value):
        if key and value and not key.lower().startswith('pwbus_'):
            self.response[key] = value

    """ REGISTRY METHODS """

    # _Flow.getRegistry
    #
    def getRegistry(self):
        return self.registry

    # _Flow.setRegistry
    #
    def setRegistry(self, registry):
        self.registry = registry

    # _Flow.getSendTo
    #
    def getSendTo(self):
        return self.registry["flow.out.resource_name"]

    # _Flow.getRecvFrom
    #
    def getRecvFrom(self):
        return self.registry["flow.in.resource_name"]

    # _Flow.isPseudoSyncInEnabled
    #
    def isPseudoSyncInEnabled(self):
        return self.registry["flow.in.pseudosync_mode"] \
            if "flow.in.pseudosync_mode" in self.registry else True

    # _Flow.isPseudoSyncOutEnabled
    #
    def isPseudoSyncOutEnabled(self):
        return self.registry["flow.out.pseudosync_mode"] \
            if "flow.out.pseudosync_mode" in self.registry else True

    # _Flow.isTaskEnabled
    #
    def isTaskEnabled(self):
        return self.registry["flow.execute_tasks"] \
            if "flow.execute_tasks" in self.registry else True

    # def getWaitBackendResponse(self):
    #     if not self.wait_backend_response:
    #         self.wait_backend_response = \
    #             float(self.registry['flow.response_delay_ms']) \
    #             if 'flow.response_delay_ms' in self.registry else 0
    #     return self.wait_backend_response

    # _Flow.getTTL - message out expire time in seconds (default=10s)
    #
    def getTTL(self):
        return int(self.registry["flow.out.ttl_seconds"]
                   if self.registry["flow.out.ttl_seconds"] else 10)

    # _Flow.isDebugEnabled
    #
    def isDebugEnabled(self):
        return self.registry['engine.debug']

    # _Flow.getChannelName
    #
    def getChannelName(self):
        return self.registry['channel']

    # _Flow.stop
    #
    def stop(self):
        self.running = False

    # _Flow.stop
    #
    def isRunning(self):
        return self.running

# FlowIn
#
#


class FlowIn(_Flow):

    def __init__(self, flow_type, connector):
        super().__init__(flow_type, connector)

    # FlowIn.getPayloadFormat
    #
    def getPayloadFormat(self):
        return self.registry["flow.in.payload_format"]

    # FlowOut.getHost
    #
    def getHost(self):
        return self.registry["flow.in.host"] if "flow.in.host" in self.registry else None

    # FlowOut.getPort
    #
    def getPort(self):
        return self.registry["flow.in.port"] if "flow.in.port" in self.registry else None

    # FlowIn.prepareMessage
    #
    def prepareMessage(self, payload):
        # Move payload headers entries to Headers
        for field in list(payload):
            if field.startswith('Pwbus-'):
                self.setHeadersEntry(field, payload[field])
                del payload[field]

        # set payload
        message = self.getMessage()
        message.setPayloadFormat(self.registry["flow.in.payload_format"])
        message.setPayload(payload)

        headers = message.getHeaders()
        message.setMessageId(secrets.token_urlsafe())
        message.setHeadersEntry('Pwbus-Message-Id', message.getMessageId())
        if not 'Pwbus-Correlation-Id' in headers:
            message.setCorrelationId(secrets.token_urlsafe())
            message.setHeadersEntry(
                'Pwbus-Correlation-Id', message.getCorrelationId())
        task_id = headers['Pwbus-Task-Id'] \
            if 'Pwbus-Task-Id' in headers else None
        message.setTaskId(task_id)

        # set headers
        message.setHeadersEntry("Pwbus-Status-Code", 200)
        self.setMessage(message)

# FlowOut
#
#


class FlowOut(_Flow):

    def __init__(self, flow_type, connector):
        super().__init__(flow_type, connector)

    # FlowOut.getPayloadFormat
    #
    def getPayloadFormat(self):
        return self.registry["flow.out.payload_format"]

    # FlowOut.getHost
    #
    def getHost(self):
        return self.registry["flow.out.host"] if "flow.out.host" in self.registry else None

    # FlowOut.getPort
    #
    def getPort(self):
        return self.registry["flow.out.port"] if "flow.out.port" in self.registry else None

    # FlowOut.prepareMessage
    #
    def prepareMessage(self, message, headers):
        self.setMessage(message)
        self.setHeaders(headers)
        self.getMessage().setCorrelationId(headers['Pwbus-Correlation-Id'])
        self.getMessage().setMessageId(headers['Pwbus-Message-Id'])
        self.getMessage().setTaskId(
            headers['Pwbus-Task-Id'] if 'Pwbus-Task-Id' in headers else None
        )



### PWBus - Filesystem Library
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Sat Nov 16 16:36:36 -03 2019
 
import shutil

## rmdir_temp
#
def rmdir_temp(dir):
    if dir.startswith('temp'):
        return False
    try:
        shutil.rmtree(dir)
    except:
        return False

    return True




# PWBus - RedisEngineThread Class
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Sat Nov 16 16:36:36 -03 2019

import traceback

from pwbus.commons.logging import *
from pwbus.engines._engine_thread import _EngineThread


class RedisEngineThread(_EngineThread):

    def __init__(self, thread_id, name, channel_registry):
        super().__init__(thread_id, name, channel_registry)



# PWBus - _TransformJSON Class
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Mon Nov 18 07:36:30 -03 2019

import traceback

from pwbus.commons.logging import *
from pwbus.transformations._transform import _Transform


class _TransformJSON(_Transform):

    # _TransformJSON.parsein - return Dict
    #
    def parse_in(self, payload):
        return self.stringToJson(data=payload)

    # _TransformJSON.parse_out - receive Dict
    #
    def parse_out(self, payload):
        return self.jsonToString(data=payload)




# PWBus - RedisFlowIn & RedisFlowOut Classes
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Sat Nov 16 16:36:36 -03 2019

from time import sleep
import traceback

from pwbus.commons.logging import *
from pwbus.flows._flow import FlowIn, FlowOut
from pwbus.connectors.redis import RedisConnectorIn, RedisConnectorOut

# RedisFlowIn
#
#


class RedisFlowIn(FlowIn):

    def __init__(self, request=None, host=None, port=None):
        super().__init__('in', RedisConnectorIn(host=host, port=port))
        self.host = host
        self.port = port

    # RedisFlowIn.receive
    #
    def receive(self, timeout=300):
        DEBUG = self.isDebugEnabled()
        resource_name = self.getRecvFrom()

        if DEBUG:
            log_debug(
                f'RedisFlowIn.receive - Started to [{resource_name}] format={self.getPayloadFormat()}')

        while True:
            try:
                if DEBUG:
                    log_debug(
                        "RedisFlowIn.receive.message - Getting connection")

                with self.getConnector().getConnectionPool().item() as connection:

                    # pooling for wait messages
                    while True:

                        # for pseudosync tasks, with SADD, SPOP and SSCAN (client)
                        if self.isPseudoSyncInEnabled():

                            while True:

                                if DEBUG:
                                    log_debug(
                                        f'⏱️  RedisFlowIn.receive.message - Waiting for messages from [{resource_name}|PSEUDOSYNC] ...')

                                # pattern event notification - for SADD command
                                # (https://redis.io/commands/blpop#pattern-e
                                tuple_in = connection.spop(resource_name)

                                if tuple_in:

                                    # convert redis payload from tuple type
                                    payload_in = ''.join(
                                        tuple_in.decode('utf-8'))
                                    if DEBUG:
                                        log_debug(
                                            f'↘️  RedisFlowIn.receive.message - Message received successfully from [{resource_name}|PSEUDOSYNC] - payload_in [{payload_in}]')
                                    return payload_in

                                connection.brpop(
                                    f'{resource_name}.helper', timeout)

                        # for async tasks, with LPUSH and BRPOPLPUSH
                        else:
                            if DEBUG:
                                log_debug(
                                    f'⏱️  RedisFlowIn.receive.message - Waiting for messages from [{resource_name}] ...')
                            tuple_in = connection.brpoplpush(
                                resource_name,
                                name + '_reliable',
                                timeout
                            )

                            if tuple_in:
                                # convert redis payload from tuple type
                                payload_in = ''.join(tuple_in.decode('utf-8'))
                                if DEBUG:
                                    log_debug(
                                        f'↘️  RedisFlowIn.receive.message - Message received successfully from [{resource_name}] - payload_in [{payload_in}]')
                                return payload_in

            except KeyboardInterrupt:
                log_debug('RedisFlowIn.receive.message - Interrupted by user')
                self.stop()

            except:
                log_error(
                    traceback, f'RedisFlowIn.receive.message - Failed on receive message from [{self.getRecvFrom()}]')
                sleep(5)


# RedisFlowOut
#
#
class RedisFlowOut(FlowOut):

    def __init__(self, host=None, port=None):
        super().__init__('out', RedisConnectorOut(host=host, port=port))
        self.host = host
        self.port = port

    # RedisFlowOut.send
    #
    def send(self):
        DEBUG = self.isDebugEnabled()

        resource_name = self.getSendTo()
        payload_out = self.getMessage().getPayloadAsString()

        if DEBUG:
            log_debug(
                f'RedisFlowOut.send - Started to [{resource_name}] - payload_out [{payload_out}]')

        try:
            with self.getConnector().getConnectionPool().item() as connection:

                if self.isPseudoSyncOutEnabled():
                    position = connection.sadd(resource_name, payload_out)
                    if position:
                        if DEBUG:
                            log_debug(
                                f'↗️  RedisFlowOut.send - Redis SADD in position [{position}]')

                        # pattern event notification
                        # (https://redis.io/commands/blpop#pattern-event-notification)
                        connection.lpush(f'{resource_name}.helper', '1')
                        return True
                    else:
                        log_debug(
                            f'RedisFlowOut.send - Redis SADD failed, payload already exists in [{resource_name}]')
                else:
                    position = connection.lpush(resource_name, payload_out)

                    if position:
                        if DEBUG:
                            log_debug(
                                f'↗️  RedisFlowOut.send - Redis LPUSH in position [{position}]')
                        self.setResponseEntry("queue_position", position)
                        return True
                    else:
                        log_debug(
                            f'RedisFlowOut.send - Redis LPUSH failed to [{resource_name}]')
                        return False

        except:
            log_error(
                traceback, f'RedisFlowOut.send - Failed to send the message to [{resource_name}]')

        return False



# PWBus - Task Class
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Mon Nov 18 10:58:09 -03 2019

import importlib
from time import sleep
from datetime import datetime
import traceback

from pwbus.commons.logging import *
from pwbus.tasks.task_unique_id import TaskUniqueId

# Task
#
#


class Task():

    def __init__(self, payload):
        self.payload = payload
        self.setField('__pwbus-task-unique-id__', TaskUniqueId().get())

    # Task.execute
    #
    def execute(self):
        self.main()
        return self.payload

    # Task.main - override this in your task
    #
    def main(self):
        return None

    # Task.setField
    #
    def setField(self, name, value):
        self.payload[name] = value

    # Task.getField
    #
    def getField(self, name):
        return self.payload[name]

    # Task.deleteField
    #
    def deleteField(self, name):
        del self.payload[name]

    # Task.getDateTime
    #
    def getDateTime(self):
        return datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f')

    # Task.getDateTimeFormat
    #
    def getDateTimeFormat(self, my_format):
        return datetime.now().strftime(my_format)

    # Task.sleep
    #
    def sleep(self, ms):
        sleep(ms)



# PWBus - SqsEngineThread Class
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Sat Dec  7 09:46:09 -03 2019

import traceback

from pwbus.commons.logging import *
from pwbus.engines._engine_thread import _EngineThread


class SqsEngineThread(_EngineThread):

    def __init__(self, thread_id, name, channel_registry):
        super().__init__(thread_id, name, channel_registry)



# PWBus - Logging library
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Sat Nov 16 08:49:20 -03 2019

from datetime import datetime
from time import time
from threading import current_thread
from os import environ

from pwbus.engines.engine_monitor_event import EngineMonitorEvent
from pwbus.commons.cprint import cprint, cprint_rainbow
from sys import stdout, exit
import random

LOG_DEBUG = True
LOG_ERROR = True
date_time = None
debug_color = 'green'

# logging.set_debug
#
def set_debug(enabled):
    global LOG_DEBUG
    LOG_DEBUG = enabled

# logging.set_debug_color
#
def set_debug_color(color):
    global debug_color
    debug_color = color


# logging.print_datetime
#
def print_datetime(arg1, arg2=''):
    global date_time
    global debug_color
    now = datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")
    this_thread = current_thread().getName()
    this_thread = "{:<10}".format(this_thread)

    if not date_time or now[0:9] != date_time[0:9]:
        date_time = now
        print(this_thread, now, '|-------------------------------------------------------------------')

    try:
        print(f'{this_thread[0:10]} {now[10:]} ', end='')
        if environ['ENVIRONMENT'] == 'production':
            print(f'{arg1} {arg2}')
        else:
            cprint(arg1, color='bright_white')
            cprint_rainbow(f' {arg2}', key=this_thread[0:10])
            print()
    except:
        cprint(f'{this_thread[0:10]} {now[10:]} ', color='bright_black')
        cprint(arg1, color='bright_white')
        cprint(f' {arg2}', color=debug_color)
        print()

# logging.getMillis
#


def getMillis():
    return int(round(time() * 1000))


# logging.log_debug
#
def log_debug_success(arg1, arg2=''):
    global LOG_DEBUG
    if LOG_DEBUG:
        print_datetime('[GREAT] 👍', arg1)
        if arg2 != '':
            print_datetime('       👍', arg2)
        print("\n")
        stdout.flush()


# logging.log_debug
#
def log_debug(arg1, arg2=''):
    global LOG_DEBUG
    if LOG_DEBUG:
        print_datetime('[DEBUG]::', arg1)
        if arg2 != '':
            print_datetime('       ::', arg2)
        stdout.flush()


# logging.log_debug_var
#
def log_debug_var(key, value, arg1=''):
    global LOG_DEBUG
    if LOG_DEBUG:
        print_datetime(f'[DEBUG]>> {key} = [{value}]')
        if arg1 != '':
            print_datetime('       >>', arg1)
        stdout.flush()

# logging.log_error
#


def log_error(traceback, message):
    global LOG_ERROR
    EngineMonitorEvent().pushEvent(
        {"type": "error", "message": message, "exception": traceback.format_exc()})
    if LOG_DEBUG:
        formatted_lines = traceback.format_exc().splitlines()
        print_datetime(
            f'[ERROR]## 🟥 {message}\n       ## Message: {formatted_lines[-1]}')
        traceback.print_exc()
        stdout.flush()

# logging.log_fatal
#


def log_fatal(message):
    EngineMonitorEvent().pushEvent(
        {"type": "fatal", "message": message, "exception": ""})
    print_datetime('[FATAL]!! 🟥 Exiting with fatal error -', message)
    exit()


# logging.log_message_dump
#
def log_message_dump(message, message_dump):
    global LOG_DEBUG
    if LOG_DEBUG:
        if message_dump:
            print_datetime(f'[DUMP] .. {message} - message: {message_dump}')
        else:
            print_datetime(f'[DUMP] .. {message} - empty message')
        stdout.flush()

# logging.log_exit
#


def log_exit(arg1):
    print_datetime('[EXIT] !! 🟥 ', arg1)
    exit()

# logging.log_warn
#


def log_warn(arg1, arg2=''):
    global LOG_DEBUG
    if LOG_DEBUG:
        print_datetime('[WARN]  :: ⚠️ ', arg1)
        if arg2 != '':
            print_datetime('       ::', arg2)
        stdout.flush()



# PWBus - EngineMonitor Class
#:
#:  maintainer: fabio.szostak@perfweb.com.br | Mon Dec 16 13:05:27 -03 2019

from threading import Thread
import asyncio
from datetime import datetime
import websockets
from json import dumps
import traceback

from pwbus.commons.logging import *
from pwbus.connectors.redis import RedisConnector
from pwbus.engines.engine_monitor_event import EngineMonitorEvent

# EngineMonitor - Get engine monitor channel settings
#
#


class EngineMonitor(Thread, RedisConnector):
    def __init__(self):
        Thread.__init__(self)
        RedisConnector.__init__(self)

    async def monitor(self, websocket, path):
        try:
            with self.getConnectionPool().item() as connection:
                event = EngineMonitorEvent(connection)
                while True:
                    data = {
                        "pwbus_task_unique_id": ''.join(connection.get("pwbus_unique").decode('utf-8')),
                        "pwbus_event": event.fetchEvent(),
                        "pwbus_flow_success": event.getData("pwbus_flow_success"),
                        "pwbus_flow_errors": event.getData("pwbus_flow_errors"),
                        "pwbus_transformation_errors": event.getData("pwbus_transformation_errors"),
                        "pwbus_task_errors": event.getData("pwbus_task_errors")
                    }
                    await websocket.send(dumps(data))
                    await asyncio.sleep(2)
        except:
            pass

    # EngineMonitor.run
    #
    def run(self):
        try:
            with self.getConnectionPool().item() as connection:
                event = EngineMonitorEvent(connection)
                event.pushEvent(
                    {"type": "info", "message": "PWBus Started", "exception": ""})
                event.setValue("pwbus_flow_success", 0)
                event.setValue("pwbus_flow_errors", 0)
                event.setValue("pwbus_transformation_errors", 0)
                event.setValue("pwbus_task_errors", 0)

            # Start server monitor
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            start_server = websockets.serve(self.monitor, "0.0.0.0", 8888)
            asyncio.get_event_loop().run_until_complete(start_server)
            log_debug('👁️  EngineMonitor.run - Server monitor started')
            asyncio.get_event_loop().run_forever()

        except:
            pass
